/*
 * business-options-gnome.h -- Initialize the Business Options
 *
 * Written By: Derek Atkins <warlord@MIT.EDU>
 * Copyright (C) 2002 Derek Atkins
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 51 Franklin Street, Fifth Floor    Fax:    +1-617-542-2652
 * Boston, MA  02110-1301,  USA       gnu@gnu.org
 */
/** @addtogroup GUI
    @{ */
/** @addtogroup GuiOptions Options Dialog
    @{ */

#ifndef GNC_BUSINESS_OPTIONS_H_
#define GNC_BUSINESS_OPTIONS_H_

/**
 * Set up the business and counters pages in the File Preferences dialog.
 */
#ifdef __cplusplus
extern "C"
{
#endif

void gnc_business_options_gnome_initialize(void);

#ifdef __cplusplus
}
#endif
#endif /* GNC_BUSINESS_OPTIONS_H_ */
/** @}
    @} */
